/***************************************************************************\
| Channel.cls                                                               |
| ~~~~~~~~~~~                                                               |
|    A class (to be invoked through ::Requires) to implement a connection   |
|    to a TCP/IP session with a client or server.                           |
|    Brian {Hamilton Kelly}                                                 |
|                                                                           |
|  Last modified: 25-Nov-1999                                               |
\***************************************************************************/
channelVersion = 2.00
/***************************************************************************\
|  Change history                                                           |
|***************************************************************************|
|  Version  Date         Reason                                             |
|  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  |
|  1.00   25-Nov-1999    Split out from embedding in my Nntpd.cmd           |
|  2.00   16-May-2001    Allow dynamic suppression of debugging             |
|                                                                           |
\***************************************************************************/

/***************************************************************************\
|                                                                           |
|     CLASS Channel                                                         |
|                                                                           |
|        Parameters:                                                        |
|            Socket - the socket allocated to the incoming connection       |  
|            Logger - stream allocated to log via syslog daemon             |
|            Debug  - whether debugging is to take place                    |
|                                                                           |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/

::Class Channel Public

   ::Method init
      expose Socket Buffer CrLf Logger Debug origDebug
      use arg Socket, Logger, Debug
      origDebug = Debug /* Cannot turn on if never on, but can be off */
      Buffer = ""       /* We've received nothing as yet */
      CrLf = "0D0A"x    /* Preset constant used elsewhere */

   /************************************************************************\
   |                                                                        |
   |  METHOD  Channel~Socket                                                |
   |                                                                        |
   |      Parameters:                                                       |
   |          (none)                                                        |
   |      Returns:                                                          |
   |          The Socket being used for the Channel                         |
   |      Notes:                                                            |
   |                                                                        |
   \************************************************************************/

   ::Method socket      /* What socket are we using? */
      expose Socket
      return Socket

   /************************************************************************\
   |                                                                        |
   |  METHOD  Channel~Logger                                                |
   |                                                                        |
   |      Parameters:                                                       |
   |          (none)                                                        |
   |      Returns:                                                          |
   |          The Logger object being used for writing to the syslog daemon |
   |      Notes:                                                            |
   |                                                                        |
   \************************************************************************/

   ::Method logger      /* What syslog route are we using? */
      expose Logger
      return Logger

   /************************************************************************\
   |                                                                        |
   |  METHOD  Channel~Read                                                  |
   |                                                                        |
   |      Parameters:                                                       |
   |          [timeout] - a value in seconds; if exceeded, read is abandoned|
   |      Returns:                                                          |
   |          Zero for successful read, -2 for timeout, -1 for socket error |
   |      Notes:                                                            |
   |          The input received is left for retrieval by Channel~Input     |
   |          If Debug is enabled, incoming text is logged                  |
   |                                                                        |
   \************************************************************************/

   ::Method Read
      expose Buffer CrLf Socket Logger CommandLine Debug
      use arg timeout

      argcnt = arg()
      if argcnt > 0 & timeout~Datatype("W") then
         argcnt = 0        /* Ignore malformed "timeout" parameter */

      do while Buffer~Pos( CrLf ) = 0
         if argcnt > 0 then do   /* Read with timeout */
            r.0 = 1; r.1 = Socket
            BytesRcvd = SockSelect( "r.", "", "", timeout )
            if BytesRcvd > 0 then
               BytesRcvd = SockRecv( Socket, "Traffic", 1024 )
            else
               BytesRcvd = -2
         end
         else     /* Read without any timeout */
            BytesRcvd = SockRecv( Socket, "Traffic", 1024 )

         if BytesRcvd <= 0 then do        /* Remote socket has vanished */
            if BytesRcvd = -2 then        /* No input within timeout */
               answer = -2
            else if BytesRcvd < 0 then do /* Receive problem */
               Logger~Send( "LOG_WARNING", "Receive failure:" errno )
               answer = -1
            end
            else do                       /* Socket has closed */
               Logger~Send( "LOG_INFO", "Remote has cleared:" errno )
               answer = -1
            end
            Traffic = CrLf
         end
         else
            answer = 0

         Buffer = Buffer || Traffic    /* Append input to bufferful */
      end      /* Repeat until a CrLf is present in stream */

      split = Buffer~Pos( CrLf )
      CommandLine = Buffer~Left( split-1 )
      Buffer = Buffer~Substr( split+2 )
      if Debug then
         Logger~Send( "LOG_DEBUG", "<<<"CommandLine )
      return answer

   /************************************************************************\
   |                                                                        |
   |  METHOD  Channel~Input                                                 |
   |                                                                        |
   |      Parameters:                                                       |
   |          (none)                                                        |
   |      Returns:                                                          |
   |          The dataline just retrieved from the channel                  |
   |      Notes:                                                            |
   |                                                                        |
   \************************************************************************/

   ::Method Input
      expose CommandLine
      return CommandLine

   /************************************************************************\
   |                                                                        |
   |  METHOD  Channel~Write                                                 |
   |                                                                        |
   |      Parameters:                                                       |
   |          Text: The data to be sent to the channel                      |
   |      Returns:                                                          |
   |          Zero if successful, otherwise RxSock reason for problem       |
   |      Notes:                                                            |
   |          Appends a CrLf if not already present in Text                 |
   |          If Debug is defined, logs outgoing traffic to syslog          |
   |                                                                        |
   \************************************************************************/

   ::Method Write
      expose Debug Socket CrLf Logger
      use arg Text
      if Debug then
         Logger~Send( "LOG_DEBUG", ">>>"Text )

      if Text~Right(2) = CrLf then
         Text = Text || CrLf

      BytesSent = 0
      do while BytesSent < Text~Length
         BytesSent = SockSend( Socket, Text )
         if BytesSent < 0 then do
            return errno
         end
         if BytesSent < Text~Length then do
            Text = Text~Substr( BytesSent+1 )
            BytesSent = 0
         end
      end
      return 0

   /************************************************************************\
   |                                                                        |
   |  METHOD  Channel~WriteGuarded                                          |
   |                                                                        |
   |      Parameters:                                                       |
   |          Text: Material to be written to channel                       |
   |      Returns:                                                          |
   |          via Channel~Write, so the latter's result codes               |
   |      Notes:                                                            |
   |          Prepends a '.' to each line if it starts with '.'             |
   |                                                                        |
   |                                                                        |
   \************************************************************************/

   ::Method WriteGuarded
      expose CrLf
      use arg Text

      /* We add an explicit trailing CrLf, if needed, to avoid potential */
      /* complications if Channel~Write were to try to do so as well     */
      if Text~Right(2) = CrLf then
         Text = Text || CrLf

      EoL = Text~Pos( CrLf )
      do while EoL > 0              /* Guaranteed to happen at least once */
         Portion = Text~Left( EoL+1 )  /* Include the explicit CrLf */
         Text = Text~Substr( EoL+2 )
         if Portion~Left(1) = '.' then
            Portion = '.' || Portion
         answer = self~Write( Portion )
         if answer = 0 then
            Return answer
         EoL = Text~Pos( CrLf )
      end
      Return 0    /* The final explicit (or added) CrLf has been written */


   /************************************************************************\
   |                                                                        |
   |  METHOD  Channel~Debug                                                 |
   |                                                                        |
   |      Parameters:                                                       |
   |          debug: New state of Debug flag (0 or 1)                       |
   |      Notes:                                                            |
   |          Used to cancel debugging temporarily (when massive data is    |
   |          expected).  CANNOT initiate debugging if never enabled        |
   |                                                                        |
   \************************************************************************/

   ::Method Debug
      expose Debug origDebug
      use arg newDebug
      if origDebug then Debug = newDebug
      return