/***************************************************************************\
| Pop3d.cmd                                                                 |
| ~~~~~~~~~                                                                 |
| A daemon to handle incoming requests for POP3 service, and to create a    |
| server for each request.  These servers exist as separate threads of the  |
| main process, which is far more efficient than creating a separate        |
| /process/ for each incoming connection.                                   |
|                                                                           |
| Based, in part on:                                                        |
|    Alex Chapman's <alex@budgetweb.com> NEWNEWS.CMD, GETMAIL.CMD and       |
|        other parts of his KA9Q-compatible WARP DIS package (v2.1);        |
|    Config file reading from Albert Crosby's RexxNews.cmd, which was       |
|        an early NNTP client;                                              |
|    Kurt Maerker's <Kurt_Maerker@vnet.ibm.com> examples of NetCentric      |
|        Computing with Object Rexx                                         |
| and developed from my own SMTPD.cmd                                       |
|                                                                           |
|  Last modified: 18-May-2001                                               |
\***************************************************************************/
   VerNumb = 0.25
/***************************************************************************\
|  Change history                                                           |
|***************************************************************************|
|  Version  Date         Reason                                             |
|  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  |
|  0.10   21-Jan-2001    First attempt developed by editing my Smtpd        |
|  0.11   21-Jan-2001    Handle the QUIT command properly; rewriting mail   |
|  0.12   21-Jan-2001    Handle the TOP and RETR commands                   |
|  0.13   22-Jan-2001    Corrected format of response to STAT command       |
|  0.14   04-Feb-2001    Convert dividing "From " into Return-Path          |
|  0.15   04-Feb-2001    Corrected number of lines returned by TOP command  |
|  0.16   11-Feb-2001    Deal with malformed ::MailBox object on disk       |
|  0.17   20-Feb-2001    Removed forgotten SAYs from debugging              |
|  0.18   20-Feb-2001    Dealt (?) with mailbox remaining locked on cleanup |
|  0.19   19-Mar-2001    Was possible for "inheaders" to appear undefined   |
|  0.20   23-Feb-2001    Open Configuration file permitting sharing         |
|  0.21   27-Mar-2001    Split debug down to trace or debug                 |
|  0.22   16-May-2001    Don't unload utility & sockets libraries on exit   |
|  0.23   16-May-2001    Correctly handle closing of an empty mailbox       |
|  0.24   17-May-2001    Corrected problem from 27-Mar tracing trick        |
|  0.25   18-May-2001    Remove extraneous debugging printout               |
|                                                                           |
\***************************************************************************/

parse source . . filename

/***************************************************************************\
|                                                                           |
|  The next command will fail if the Y2K fixes for REXX aren't applied      |
|                                                                           |
\***************************************************************************/

settings.FileDate = Stream(filename,'C','QUERY TIMESTAMP')

/* Ensure that we're running under ObjRexx */

parse version processor versionum versiondate
if Translate( processor ) <> "OBJREXX" then do
   SAY "Sorry, this program *requires* to be run under Object REXX."
   SAY "From an OS/2 command prompt, type SWITCHRX, follow the"
   SAY "instructions, and reboot.  All existing valid Classic Rexx"
   SAY "programs should run under ORexx too."
   Exit 999
end

settings.Version = "Rexx POP3 daemon v"VerNumb "by B{HK}. ("settings.FileDate")"

settings.VarNames = "",    /* List of valid options, and their types */
                    "Version              -",
                    "debug                B",
                    "trace                B",
                    "Pop3dDir             -",
                    "EtcDir               -",
                    "hostname             S",
                    "port                 W",
                    "ka9qPath             S",
                    "MQueue               S",  /* Unused, but in [common] */
                    ""

SIGNAL on HALT name ShutDown

/* Following need initialization, since ShutDown tries to close them */
Socket = -1
ClientSocket = -1

if Prelude( filename ) < 0 then   /* Load various DLLs, and config file */
   Signal ShutDown

SIGNAL ON SYNTAX name ShutDown      /* Ensure tidy up if something wrong */

if settings.debug then CALL Set     /* Display the config values */

/* The directory for Pop3 mailboxes must be defined */
if Defined("ka9qPath") then do
      SAY "Mailbox directory undefined"
      Signal ShutDown
end

hosts_file = SysSearchPath( 'ETC', 'HOSTS' )

/* Discover the hostname for this server */
if Defined("hostname") then 
   destination = settings.hostname
else do
   SAY "Determining hostname from environment"
   destination = Value( "HOSTNAME", , "OS2ENVIRONMENT" )
end

if destination = '' then do
   SAY "Determining hostname from IP address"
   destination = SockGetHostID()
   if destination = '255.255.255.255' then do
      SAY 'Unable to determine local hostname'
      SAY
      SAY 'The most likely problem is that you have not configured your'
      SAY 'interface(s).  Type the following command at an OS/2 Command'
      SAY 'prompt, replacing www.xxx.yyy.zzz with your IP address, or'
      SAY "with 127.0.0.1 if you don't have a fixed IP address"
      SAY
      SAY '         ifconfig lo www.xxx.yyy.zzz'
      SAY
      SAY 'You should also include a line of the following format in your'
      SAY 'etc/hosts file ('hosts_file') for your own host.  The record should'
      SAY 'have the following format, where the long and short hostnames reflect'
      SAY 'your actual Demon subdomain and a local nickname for it'
      SAY
      SAY '              ' ip_address 'host.demon.co.uk host'
      MainLog~send( "LOG_ERR", "Host IP address undefined" )
      Signal ShutDown
   end
   retcode = SockGetHostByAddr( destination, 'host.!' )
   if retcode < 0 then do
      SAY 'SockGetHostByAddr()' errno
      MainLog~send( "LOG_ERR", "IP address lookup failure" errno )
      Signal ShutDown
   end
   ip_address = destination
   Parse Lower var host.!name destination
end
else do      /* Hostname was in environment */
   SAY "Finding IP number for" destination
   retcode = SockGetHostByName( destination, 'host.!' )
   if retcode < 0 then do
      SAY 'SockGetHostByAddr()' errno
      MainLog~send( "LOG_ERR", "IP address lookup failure" errno )
      Signal ShutDown
   end
   ip_address = host.!addr
   Parse Lower var host.!name destination
end

MainLog~send( "LOG_INFO", "pop3d hostname" destination "["ip_address"]" )
SAY "pop3d hostname" destination "["ip_address"]"

if destination = 'HOST.!NAME' then do
   SAY 'Unable to determine local hostname'
   SAY
   SAY 'The most likely cause is that you have not included a line in your'
   SAY 'etc/hosts file ('hosts_file') for your own host.  The record should'
   SAY 'have the following format, where the long and short hostnames reflect'
   SAY 'your actual Demon subdomain and a local nickname for it'
   SAY
   SAY '              ' ip_address 'host.demon.co.uk host'
   MainLog~send( "LOG_ERR", "Pop3d: No local name for server host" )
   Signal ShutDown
end

settings.hostname = destination

if settings.debug then do
   if settings.loadUtil then SAY "Utilities loaded"
   if settings.loadSock then SAY "Sockets loaded"
   if settings.loadUPM then SAY "LAN Server utilities loaded"
end

.local~Pop3d.settings = settings.

/***************************************************************************\
|                                                                           |
|  This is where we start work;  Open socket at port 110                    |
|                                                                           |
\***************************************************************************/

Socket = ListenPort( settings.port )
if Socket < 0 then
   Signal ShutDown

/* Close the server when program is interrupted */
Signal on Halt name ShutDown

/***************************************************************************\
|                                                                           |
|  Wait for client(s) to connect; spawn a handler for each                  |
|                                                                           |
\***************************************************************************/

SAY "Waiting for clients to connect; press Ctrl/C to exit"
do forever
   ClientSocket = SockAccept( Socket )
   MainLog~Send( "Log_INFO", "Client connected on socket",
                                  ClientSocket"; starting handler" )
   servit = .server~new    /* Create a new server process for it */
   rs = servit~startup( ClientSocket )              /* Spawn sub-thread */
   MainLog~Send( "LOG_NOTICE", rs )
end
/* Fall through to exit the program */

/***************************************************************************\
|                                                                           |
|  GOTO here when troubles, or on being interrupted; clean up & shut down   |
|                                                                           |
\***************************************************************************/

ShutDown:
MainLog~Send( "LOG_INFO", "Server exiting at" Time('Normal') )
MainLog~uninit
Call Close Socket
if ClientSocket > 0 then
   Call Close ClientSocket

/* Following "unloading" code commented out: other processes might still
   be using these libraries, so we mustn't unload them */
/*
if settings.loadUtil then do
   SAY "Unloading RexxUtil"
   CALL SysDropFuncs
   CALL RxFuncDrop 'SysLoadFuncs'
end
if settings.loadSock then do
   SAY "Unloading Sockets"
   CALL SockDropFuncs
   CALL RxFuncDrop 'SockLoadFuncs'
end
*/
if settings.loadUPM then do
   SAY "Unloading LAN Server utilities"
   CALL RxFuncDrop 'NetMisc'
end


Exit 0


/***************************************************************************\
|                                                                           |
|     PROCEDURE ListenPort                                                  |
|                                                                           |
|        Parameters:                                                        |
|               Port: Port number to be created on which to listen          |
|                                                                           |
|        Returns:                                                           |
|               Socket number created, or -1 for errors                     |
|                                                                           |
|        Notes:                                                             |
|               Access the MainLog syslog logging object                    |
|                                                                           |
\***************************************************************************/

ListenPort: PROCEDURE expose MainLog
   parse arg Port
   /* Create a TCP socket */
   Socket = SockSocket( "AF_INET", "SOCK_STREAM", "0" )
   if Socket < 0 then do
      reason = errno
      MainLog~Send( "LOG_ERR", "Cannot create listener socket. Reason:" reason)
      Return -1
   end
   /* Find our local IP address and bind socket to that */
   /* Host.!addr = SockGetHostId()    **OR** */
   /* Don't bind socket to any specific address */
   Host.!addr = 'INADDR_ANY'
   Host.!family = "AF_INET"
   Host.!port = Port
   rc = SockBind( Socket, "Host.!" )
   if rc < 0 then do
      reason = errno
      MainLog~Send( "LOG_ERR", "Unable to bind to port" Port "Reason:" reason)
      Say "Unable to bind to port:" Port "Reason:" reason
      Call Close Socket
      Return -1
   end
   /* Listen at the port, allow five clients in queue */
   rc = SockListen( Socket, 5 )
   if rc < 0 then do
      reason = errno
      Say "Unable to listen at port:" Port "Reason:" reason
      MainLog~Send( "LOG_ERR", "Unable to listen at port" Port "Reason:" reason)
      Call Close Socket
      Return -1
   end

   Return Socket

/***************************************************************************\
|                                                                           |
|  PROCEDURE Prelude                                                        |
|                                                                           |
|        Parameters:                                                        |
|          name: Full path and filename for the currently-executing program |
|                                                                           |
|        Returns:                                                           |
|               Zero on successful completion, or -1 for errors             |
|                                                                           |
|        Notes:                                                             |
|          Loads utility & other libraries as required                      |
|          Sets some defaults, which can be overridden in config file       |
|          Reads configuration from file & stores in global stem settings.  |
|          Creates & uses the global MainLog syslog logging object          |
|          NB Uses the same (SMTP.CFG) file as the SMTP client & server     |
|                                                                           |
\***************************************************************************/

Prelude: PROCEDURE expose MainLog settings. 
   use arg name

   settings.loadUtil = 0
   settings.loadSock = 0
   settings.loadUPM  = 0

   if RxFuncQuery('SysLoadFuncs') then
   do
      settings.loadUtil = 1     /* Remember this so that it can be unloaded */
      xrc = RxFuncAdd('SysLoadFuncs','RexxUtil','SysLoadFuncs')
      xrc = SysLoadFuncs()
   end

   if RxFuncQuery( 'SockLoadFuncs' ) then do
      settings.loadSock = 1
      xrc = RxFuncAdd( 'SockLoadFuncs', 'RXSOCK', 'SockLoadFuncs' )
      xrc = SockLoadFuncs()
   end

   if RxFuncQuery( 'NetMisc') then do
      settings.loadUPM = 1 /* LAN Server utilities allow user/pass checking */
      xrc = RxFuncAdd( 'NetMisc', 'LsRxUt', 'NetMisc' )
      if xrc <> 0 | RxFuncQuery( 'NetMisc' ) then do
         SAY "This program requires access to OS/2 User Profile Management."
         SAY "This can be achieved through the LAN Server REXX APM utilities,"
         SAY "which can be downloaded from IBM's website at:"
         SAY "<URL:http://www-4.ibm.com/software/os/warp/downloads/e1a1/",
                  || "wslsrxut.zip>"
         return -1
      end
   end /* do */

   settings.EtcDir   = value('ETC',,'OS2ENVIRONMENT')
   
   settings.debug = 0
   settings.trace = 0
   settings.port = 110
   
   /* Create an object through which to send messages to syslogd */
   MainLog = .Syslog~new( "Pop3d", "LOG_DAEMON" )
   MainLog~Send( 'LOG_INFO', settings.Version || " started." )

   /* Knowing where *we* are, look for the configuration file */
   settings.Pop3dDir = filespec('drive',name)||filespec('path',name)
   /* In order of preference, the same directory as the program */
   CfgFilename = Stream( addslash(settings.Pop3dDir)||'SMTP.CFG','C','QUERY EXISTS')
   /* Failing that, look in the directory defined by the ETC environment */
   if CfgFileName = '' then
      CfgFilename = Stream( addslash(settings.EtcDir)||'SMTP.CFG','C','QUERY EXISTS')
   /* If that fails, perhaps there's one where we are presently? */
   if CfgFileName = '' then
      CfgFilename = Stream( 'SMTP.CFG','C','QUERY EXISTS')
   
   if LoadSettings(CfgFileName,"Pop3d") = 0 then
   do
      SAY "Unable to read config file:" CfgFileName
      Return -1
   end
   
   SAY settings.Version
   
   .local~Pop3d.settings = settings.
   return  0    /* Signal handler automatically deinstalled */

/***************************************************************************\
|                                                                           |
|  SUBROUTINE LoadSettings                                                  |
|                                                                           |
|        Parameters:                                                        |
|           CfgFile: string containing path to the initialization file      |
|           appl:    string containing name of application                  |
|                                                                           |
|                                                                           |
\***************************************************************************/

LoadSettings:   PROCEDURE expose settings.
    arg CfgFile, appl
    if CfgFile = '' then return 0
    CfgFil = .TextFile~new( CfgFile )
    Cfg = CfgFil~Read
    if Cfg = .NIL then return 0
    in_app = 0       /* Not currently in "our" section */
    had_app = 0      /* Not yet had anything of wanted */
    appl = translate(appl)
    do i=Cfg~first to Cfg~last
      if Cfg[i] <> '' & left(Cfg[i],1) <> '#' then do
         if left(Cfg[i],1) = '[' then do
            section = translate( substr( Cfg[i], 2, pos(']',Cfg[i])-2 ) )
            if appl == section | section == 'COMMON' then
               in_app = 1     /* We want this; it's "ours" */
            else
               in_app = 0     /* Ignore this; it's for someone else */
            had_app = 1
         end
         else if had_app then do
            SAY "config item: '"Cfg[i]"' not in any named section"
            return 0
         end
         else if in_app then call set Cfg[i]
      end
    end
    return 1

/***************************************************************************\
|                                                                           |
|     PROCEDURE set                                                         |
|                                                                           |
|        Parameter:                                                         |
|             a_line: string containing one line of initialization data     |
|                                                                           |
|        Notes:                                                             |
|             If called without a parameter, lists known settings           |
|                                                                           |
\***************************************************************************/

Set:  PROCEDURE expose settings.
   parse arg a_line
   if a_line <> '' then do
      parse var a_line VarName "=" VarValue
      VarName = Strip(VarName)
      VarValue = Strip(VarValue)
      i = WordPos(translate(VarName),translate(settings.VarNames))
      if i // 2 = 0 then do
         SAY 'Unknown variable:' VarName 'in config file - ignored'
         return 0
      end
      ExpectedType = Word(settings.VarNames,i+1)
      if ExpectedType = '-' then
         SAY 'Attempt to change constant' VarName '- ignored'
      else if  ExpectedType = 'S' | DataType(VarValue,ExpectedType) then
         interpret 'settings.'VarName'=VarValue'
      else
         SAY 'Improper type for' VarName '- ignored.'
      return 1
   end
   /* If called with blank parameter, lists the values in settings. */
   set.0 = Words(settings.VarNames)/2 + 4
   set.1 = settings.Version "Settings"
   set.2 = ''
   set.3 = Left("Variable",25) || Left("Value",35) || " Type"
   set.4 = Left("========",25) || Left("=====",35) || " ===="
   n = 4
   do i=1 to Words(settings.VarNames) by 2
      n = n+1
      set.n = Left(Word(settings.VarNames,i),25)
      VarName = translate("settings." || Word(settings.VarNames,i))
      interpret "if "VarName"=VarName then;",
                    "ident = '(undefined)';",
                 "else;",
                    "ident = "VarName
      set.n = set.n || Left(ident,35) || " " || Word(settings.VarNames,i+1)
   end
   do i=1 to set.0
      SAY set.i
   end
   return

/***************************************************************************\
|                                                                           |
|     SUBROUTINE Defined()                                                  |
|                                                                           |
|           Parameters:                                                     |
|              setting: name of a element of settings. array                |
|           Returns:                                                        |
|              True if element has a defined value                          |
|                                                                           |
\***************************************************************************/

Defined: PROCEDURE expose settings.
   parse arg setting
   interpret "if settings."setting "= translate('settings.'||setting) then;",
      "return 0;",
   "else;",
      "return 1;"

/***************************************************************************\
|                                                                           |
|      END OF MAIN PROGRAM & its Subroutines.  Routines & Classes follow    |
|                                                                           |
\***************************************************************************/

/***************************************************************************\
|                                                                           |
|      HERE IS WHERE WE INCLUDE EXTERNALLY-DEFINED CLASSES                  |
|                                                                           |
\***************************************************************************/

::Requires Syslog.Cls      /* Means of communication with syslog daemon */
::Requires Channel.Cls     /* TCP/IP socket communication channel */

/***************************************************************************\
|                                                                           |
|     ROUTINE  Close                                                        |
|                                                                           |
|        Parameters:                                                        |
|              Socket - the TCP/IP socket to be destroyed                   |
|                                                                           |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|              Used by a number of different classes, and main program      |
|                                                                           |
\***************************************************************************/

::Routine Close
   use arg Socket
   if Socket < 0 then
      Return
   src = SockShutDown( Socket, 2 )
   if src = 0 then do
      SAY "Error shutting socket" Socket SockSock_Errno()
      /*Return*/
   end
   src = SockClose( Socket )
   if src = 0 then
      SAY "Error closing" Socket errno
   Return



/***************************************************************************\
|                                                                           |
|     CLASS .Server                                                         |
|                                                                           |
|        Notes: Nothing need be done on instantiation                       |
|                                                                           |
\***************************************************************************/


::Class Server

/***************************************************************************\
|                                                                           |
|  METHOD  Server::startup                                                  |
|                                                                           |
|                                                                           |  
|        Parameters:                                                        |  
|            Socket - the socket allocated to the incoming connection       |  
|                                                                           |  
|        Returns:  Early, so as to spawn a sub-thread                       |  
|                                                                           |  
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/

   ::Method startup
      expose Logger settings. Socket
      use arg Socket
      settings. = .local~Pop3d.settings

      Logger = .Syslog~new( "Pop3d"||Socket, "LOG_MAIL" )
      TheChannel = .Channel~new( Socket, Logger, settings.debug )
      Call SockGetPeerName Socket, "ClientAddr.!"
      Call SockGetHostByAddr ClientAddr.!addr, "remhost.!"
      Logger~Send( "LOG_NOTICE", ,
            "Connection from '"remhost.!name"' ["ClientAddr.!addr"]" )

      response = "connection with client '"remhost.!name"' ["ClientAddr.!addr"]"
      peerinfo = remhost.!name "["ClientAddr.!addr"]"

      /* Now that we know what to say to client, send the response; the REPLY */
      /* statements return control to the calling (main) program, leaving the */
      /* remainder of this code running as a sub-thread.                      */
      theReply = "Established" response
      Welcome = "+OK" settings.hostname settings.Version

      REPLY theReply    /* Return to caller; remainder becomes new thread */
      
      timeout = 10      /* Allow only small window to be held maliciously */
      /* Send response to client */
      frc = TheChannel~Write( Welcome )
      reason = errno

      /* Create an object to handle this client's sessions */
      Processor = .CommandProcessor~new( TheChannel, peerinfo )

      if frc = 0 then
         Logger~Send( "LOG_WARNING",,
                      "Error writing socket:" frc "reason:" reason )
      else
         do forever
            status = TheChannel~Read( timeout )
            if status < 0 then
               CommandLine = status timeout
            else
               CommandLine = TheChannel~Input
            Parse var CommandLine Command Options
            Command = Command~translate   /* Upper-case it */
            Options = Options~strip       /* Remove leading/trailing spaces */
            /* Invoke relevant method to handle command */
            Message = .message~new( Processor, Command, 'I', Options )
            Message~Send            /* Activate Command */
            if Message~result = ">>>End of Session<<<" then
               Leave                /* Close down this client */
            timeout = 900    /* Increase timeout after something's been said */
         end

      /* End of client session */
      Processor~finish        /* Ensure that command processor dies */
      x = SysSleep( 1 )    /* Leave user time to read any message! */
      Logger~Send( "LOG_INFO", "Exiting at" Time( 'Normal' ) )
      Logger~uninit

      if Socket > 0 then
         Call Close Socket
      Socket = -1
      Exit        /* End of thread; let it die */

/***************************************************************************\
|                                                                           |
|     METHOD Server::Uninit                                                 |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|            Closes the client socket when server process vanishes          |
|                                                                           |
\***************************************************************************/
   ::Method Uninit
      expose Socket
      if Socket > 0 then
         CALL Close Socket
      Socket = -1
      Exit
      
/***************************************************************************\
|                                                                           |
|     CLASS TextFile                                                        |
|                                                                           |
|        Parameters:                                                        |
|            filename: File associated with the object                      |
|                                                                           |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/

::CLASS TextFile
   ::METHOD INIT
      expose filename
      use arg filename

/***************************************************************************\
|                                                                           |
|     METHOD TextFile::TimeStamp                                            |
|                                                                           |
|        Returns:                                                           |
|             Sortable timestamp of the associated file                     |
|                                                                           |
\***************************************************************************/
   ::Method TimeStamp
      expose filename
      return Stream( filename, 'C', 'QUERY TIMESTAMP' )

/***************************************************************************\
|                                                                           |
|  METHOD  TextFile::Read                                                   |
|                                                                           |
|      Returns:                                                             |
|              An array variable with the lines of the file (OR NIL)        |
|                                                                           |
\***************************************************************************/

   ::METHOD Read
      expose content filename

      if filename = '' then
         return .NIL
      TheFile = .Stream~new( filename )
      if TheFile~Open( 'Read ShareRead' ) = "READY:" then
      do
         SAY "Error opening file" filename "for reading"
         return .NIL
      end
      content = TheFile~ArrayIn( 'LINES' )
      frc = TheFile~Close
      /* Discard trailing blank line */
      if content~items > 0 then
         if content[content~last] = "" 
            then frc = content~remove( content~last ) 
      return content

/***************************************************************************\
|                                                                           |
|     CLASS CommandProcessor                                                |
|                                                                           |
|        Parameters:                                                        |
|           channel:  the duplex socket connected to the client process     |
|           peerdata: the name and IP address of the connected peer         |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/

::Class CommandProcessor

   ::Method init
      expose channel settings. peerdata hadUser mailDrop lockFile
      use arg channel, peerdata
      settings. = .local~Pop3d.settings
      response = "Starting command processor"
      channel~logger~send( "LOG_INFO", response )
      hadUser = 0
      mailDrop = .NIL
      lockFile = ""

/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::Uninit                                       |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method Uninit
      self~Finish
      
/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::Finish                                       |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method Finish
      expose lockFile mailDrop
      if mailDrop <> .NIL then
         mailDrop~Close
      if lockFile <> "" then
         CALL SysFileDelete lockFile
      mailDrop = .NIL
      lockFile = ""
         
/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::Unknown                                      |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<" (or ">>>End of Session<<<" if link lost)      |
|        Notes:                                                             |
|            Handles any unrecognized commands                              |
|                                                                           |
\***************************************************************************/
   ::Method Unknown        /* Handles all unrecognized commands */
      expose channel settings.
      use arg Command, Options

      if Command < 0 & Command~length > 0 then do  /* Socket problems */
         select
            when Command = -2 then do
               channel~logger~send( "LOG_WARNING", ,
                        "No input from client on" channel~socket"; closing" )
               channel~Write( "-ERR" settings.hostname "Timed-out after" Options[1] "seconds")
            end
            when Command = -1 then
               channel~logger~Send( "LOG_WARNING", ,
                              "Lost client on" channel~socket"; closing" )
            otherwise do
               channel~logger~Send( "LOG_NOTICE", ,
                              "Malformed command:"Command"; ignored" )
               channel~Write( "-ERR Invalid command" Command Options[1] )
               return ">>>Continue<<<"
            end
         end /* select */
         return ">>>End of Session<<<"
      end
      else do              /* Unrecognized command */
         if arg() =1 then
            Command = "UNKNOWN"
         channel~Write( "-ERR Invalid command" Command Options[1] )
         return ">>>Continue<<<"
      end

/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::Help                                         |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|        Notes:                                                             |
|            Help information on available commands (disabled)              |
|                                                                           |
\***************************************************************************/
   ::Method Help
      expose channel
      use arg Options
      hiddenCmds = .Set~Of( "UNKNOWN", "INIT", "UNINIT", "FINISH" )
      cmdSupplier = .CommandProcessor~methods( .NIL )

      channel~Write( "-ERR Invalid command: HELP" Options[1] )
      return ">>>Continue<<<"

      /* Remaining code redundant, but could be reactivated for HELP */
      channel~Write( "214-Commands supported:" )
      do while cmdSupplier~available
         cmd = cmdSupplier~index
         if hiddenCmds~hasIndex( cmd ) then
            channel~Write( "214-" || cmd )
         cmdSupplier~next
      end /* do */
      channel~Write( "214 (end)" )
      return ">>>Continue<<<"

/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::USER                                         |
|                                                                           |
|        Parameters:                                                        |
|            username provided by peer                                      |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method USER
      expose channel settings. peerdata hadUser username mailDrop
      use arg Options
      if mailDrop <> .NIL then do   /* USER/PASS already accepted */
         channel~Write( "-ERR Maildrop for user" username "already open" )
         channel~logger~send( "LOG_WARNING",, 
                                 "USER command after authentication" )
         return ">>>Continue<<<"
      end
      channel~Write( "+OK Username" Options "accepted" )
      hadUser = 1
      username = Options
      return ">>>Continue<<<"

/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::PASS                                         |
|                                                                           |
|        Parameters:                                                        |
|            password provided by user                                      |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/

   ::Method PASS
      expose expose channel settings. hadUser username mailDrop lockFile
      use arg Options
      if mailDrop <> .NIL then do   /* USER/PASS already accepted */
         channel~Write( "-ERR Maildrop for user" username "already open" )
         channel~logger~send( "LOG_WARNING",, 
                                 "PASS command after authentication" )
         return ">>>Continue<<<"
      end
      if hadUser then do
         channel~Write( "-ERR Bad sequence of commands (PASS without USER)")
         return ">>>Continue<<<"
      end

      NETUSERVALIDATE2 = 500     /* "Symbolic constant"!! */
      lsresp = NetMisc( NETUSERVALIDATE2, "info",,
                           username~Translate, Options~Translate, '' ) 

      if lsresp <> 0 | info.priv < 1 then do
         response = "-ERR Unable to access maildrop for" username,
                        "("lsresp")"
         hadUser = 0
         channel~logger~send( "LOG_ALERT",, 
               "User" username "provided invalid password" Options )
      end
      else do /* NetMisc returned zero; therefore valid authorization */
         response = "+OK"
         channel~logger~send( "LOG_INFO", "User" username "validated OK" )
         path = AddSlash( settings.ka9qPath )
         filename = username~Left(8)~Strip
         lock = path || filename || ".LCK"
         mail = path || filename || ".TXT"
         OK = 0
         if Stream( lock, 'C', 'Query Exists' ) <> "" then
            channel~logger~send( "LOG_WARNING",, 
                        "Lock file" lock "already exists" )
         else if Stream( lock, 'C', 'Open Write' ) <> "READY:" then
            channel~logger~send( "LOG_WARNING",, 
                        "Unable to create lock file" lock )
         else do
            CALL Stream lock, 'C', 'Close'
            lockFile = lock
            OK = 1
         end
         
         if OK then do
            mailDrop = .MailBox~new( mail )
            OK = mailDrop~Valid
         end
         if OK then
            response = "+OK" username "has" mailDrop~Items "messages (" ||,
                                       mailDrop~Size "octets)"
         else do
            mailDrop = .NIL
            channel~logger~Send( "LOG_ERR", "Maildrop for" username "locked" )
            response = "-ERR mailbox is already locked; retry after 15mins"
         end
      end
      channel~Write( response )
      return ">>>Continue<<<"

/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::RSET                                         |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method RSET
      expose expose channel settings. hadUser username mailDrop
      if mailDrop = .NIL then do
         channel~Write( "-ERR Invalid sequence of commands (not authorized)")
         return ">>>Continue<<<"
      end
      mailDrop~UndeleteAll
      channel~Write( "+OK" username "has" mailDrop~Items "messages (" ||,
                                    mailDrop~Size "octets)" )
      return ">>>Continue<<<"

/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::STAT                                         |
|                                                                           |
|        Parameters:                                                        |
|            (NONE)                                                         |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|                                                                           |
|        Notes:                                                             |
|            This MUST NOT return anything bar the bare details             |
|            (See RFC-1939 p.6)                                             |
|                                                                           |
\***************************************************************************/
   ::Method STAT
      expose channel mailDrop username
      if mailDrop = .NIL then do
         channel~Write( "-ERR Maildrop not opened for" username )
         return ">>>Continue<<<"
      end
      channel~Write( "+OK" mailDrop~Items mailDrop~Size )
      return ">>>Continue<<<"
      
/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::NOOP                                         |
|                                                                           |
|        Parameters:                                                        |
|            (NONE)                                                         |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|                                                                           |
\***************************************************************************/
   ::Method NOOP
      expose channel
      channel~Write( "+OK" )
      return ">>>Continue<<<"
      
/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::LIST                                         |
|                                                                           |
|        Parameters:                                                        |
|            [item]: optional item number (otherwise lists entire drop)     |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method LIST
      expose expose channel settings. mailDrop
      use arg msgNo
      if mailDrop = .NIL then do
         channel~Write( "-ERR Bad sequence of commands (no authorization)")
         return ">>>Continue<<<"
      end
      if msgNo~Datatype = "NUM" then do
         if mailDrop~Exists( msgNo ) then
            response = "-ERR No such message" msgNo
         else if mailDrop~Deleted( msgNo ) then
            response = "-ERR Message" msgNo "already deleted"
         else do
            firstLine = mailDrop~Select( msgNo )
            response = "+OK" msgNo mailDrop~Bytes
         end
         channel~Write( response )
         return ">>>Continue<<<"
      end
      else do     /* LIST entire maildrop */
         channel~Write( "+OK" mailDrop~Items "messages (" ||,
                                    mailDrop~Size "octets).  Listing follows" )
         do i = 1 to mailDrop~Items
            if mailDrop~Deleted( i ) then
               Iterate
            length = mailDrop~~Select( i )~Bytes
            channel~Write( i length )
         end
         channel~Write( "." )
         return ">>>Continue<<<"
      end
   
/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::DELE                                         |
|                                                                           |
|        Parameters:                                                        |
|            msgNo: number of message to be marked for deletion             |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method DELE
      expose channel mailDrop
      use arg msgNo
      if mailDrop = .NIL then do
         channel~Write( "-ERR Bad sequence of commands (no authorization)")
         return ">>>Continue<<<"
      end
      status = mailDrop~Delete( msgNo )
      select
         when status = -1 then do
            response = "-ERR Message" msgNo "already marked for deletion"
         end
         when status = 0 then do
            response = "-ERR No such message" msgNo
         end
         when status = 1 then do
            response = "+OK Message" msgNo "marked for deletion"
         end
      end
      channel~Write( response )
      return ">>>Continue<<<"
      
/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::TOP                                          |
|                                                                           |
|        Parameters:                                                        |
|            msgNo: the message to be retrieved                             |
|            lines: how many lines of the body to transmit                  |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method TOP
      expose channel mailDrop settings.
      use arg Options
      parse var Options msgNo lines
      if mailDrop = .NIL then do
         channel~Write( "-ERR Bad sequence of commands (no authorization)")
         return ">>>Continue<<<"
      end
      if mailDrop~Exists( msgNo ) then do
         channel~Write( "-ERR No such message" msgNo )
         return ">>>Continue<<<"
      end

      if mailDrop~Deleted( msgNo ) then do
         channel~Write( "-ERR Message" msgNo "already deleted" )
         return ">>>Continue<<<"
      end

      if lines~Datatype <> "NUM" then
         lines = 0
      if lines < 0 then
         lines = 0

      firstLine = mailDrop~Select( msgNo )
      channel~Write( "+OK Top of message" msgNo "follows" )
      if settings.trace then
         channel~Debug( 0 )   /* Cancel tracing during data body */
      inbody = .false
      do until firstLine = .NIL
         if inbody then do
            if lines = 0 then 
               Leave
            lines = lines - 1
         end
         channel~WriteGuarded( firstLine )
         if firstLine~Length = 0 then
            inbody = .true
         firstLine = mailDrop~NextLine
      end

      channel~Write( "." )
      channel~Debug( settings.debug )  /* Restart tracing */
      return ">>>Continue<<<"
      
/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::RETR                                         |
|                                                                           |
|        Parameters:                                                        |
|            msgNo: the message to be retrieved                             |
|                                                                           |
|        Returns:                                                           |
|            ">>>Continue<<<"                                               |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method RETR
      expose channel mailDrop settings.
      use arg msgNo

      if mailDrop = .NIL then do
         channel~Write( "-ERR Bad sequence of commands (no authorization)")
         return ">>>Continue<<<"
      end

      if mailDrop~Exists( msgNo ) then do
         channel~Write( "-ERR No such message" msgNo )
         return ">>>Continue<<<"
      end

      if mailDrop~Deleted( msgNo ) then do
         channel~Write( "-ERR Message" msgNo "already deleted" )
         return ">>>Continue<<<"
      end

      firstLine = mailDrop~Select( msgNo )
      channel~Write( "+OK Message" msgNo "follows" )
      if settings.trace then
         channel~Debug( 0 )   /* Cancel tracing during data body */
      do until firstLine = .NIL
         channel~WriteGuarded( firstLine )
         firstLine = mailDrop~NextLine
      end

      channel~Write( "." )
      channel~Debug( settings.debug )  /* Restart tracing */
      return ">>>Continue<<<"
      
   
/***************************************************************************\
|                                                                           |
|     METHOD CommandProcessor::Quit                                         |
|                                                                           |
|        Returns:                                                           |
|            ">>>End of Session<<<"                                         |
|                                                                           |
\***************************************************************************/
   ::Method Quit
      expose channel settings. mailDrop lockFile username
      if mailDrop <> .NIL then do
         if mailDrop~Changed then do
            template = AddSlash( settings.ka9qPath ) || "new?????.tmp"
            tempFile = SysTempFileName( template )
            if tempFile <> "" then do
               if mailDrop~Rewrite( tempFile ) then do
                  userFile = username~Left(8)~strip || ".TXT"
                  mailFile = AddSlash( settings.ka9qPath ) || userFile
                  CALL SysFileDelete mailFile
                  tempFile = tempFile~ChangeStr( '/', '\' )
                  'RENAME' tempFile userFile
               end
            end
         end
         mailDrop~Close;
         mailDrop = .NIL
      end
      if lockFile <> "" then
         CALL SysFileDelete lockFile
      lockFile = ""
      channel~Write( "+OK" settings.hostname "closing connection" )
      return ">>>End of Session<<<"

/***************************************************************************\
|                                                                           |
|     CLASS MailBox                                                         |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|                                                                           |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/

::CLASS Mailbox

/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Init                                                  |
|                                                                           |
|        Parameters:                                                        |
|            filename: full path of file containing maildrop                |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::METHOD Init
      expose maildrop. file valid isopen
      use arg filename
   
      valid = .false
      isopen = .false
      maildrop. = 0
      file = .Stream~new( filename )
      stat = file~Open( 'READ' ) 
      if stat <> "READY:" then
         return      /* Could not open file; ~Valid will be false */
         
      valid = .true
      isopen = .true
      item = 0
      starting = .false
      inheaders = .false      /* Needed to prevent race condition */
      lines = 0
      
      do while file~Chars > 0
         bytepos = file~Query( 'POSITION READ CHAR' )
         theLine = file~LineIn
         Select
   
            when theLine~Left(5) = "From " then do   /* Message separator */
               if item > 0 then do     /* Have we already got one article? */
                  maildrop.item.!last = bytepos - 2
                  maildrop.item.!size = bytepos - maildrop.item.!first
                  maildrop.item.!lines = lines
               end
               item = item + 1
               maildrop.item.!start = bytepos   /* Remember Return-Path */
               starting = .true     /* Next line starts an article */
            end
   
            when starting then do        /* Start of a new message */
               maildrop.item.!first = bytepos
               maildrop.item.!deleted = .false
               lines = 0
               inheaders = .true
               starting = .false
            end      /* Handling first line of a message */
   
            when theLine = "" then do    /* Totally blank line */
               if inheaders then do
                  maildrop.item.!body = bytepos
                  inheaders = .false
               end
            end
            
            otherwise if starting then
               lines = lines + 1
               
            
         end /* Select */
         
      end /* reading every line of the file */
      
      maildrop.!size = file~Query( 'POSITION READ CHAR' ) - 1
      maildrop.item.!last = maildrop.!size - 1
      maildrop.item.!size = maildrop.!size + 1 - maildrop.item.!first
      maildrop.item.!lines = lines
      maildrop.!items = item
      maildrop.!changed = .false

/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Uninit                                                |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Notes:                                                             |
|            Disposes of dynamic memory when daemon has finished.           |
|                                                                           |
\***************************************************************************/
   ::METHOD Uninit
      expose isopen
      self~Close

/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Close                                                 |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Notes:                                                             |
|            Ensures the mailbox file gets closed                           |
|                                                                           |
\***************************************************************************/
   ::METHOD Close
      expose maildrop. file valid isopen
      DROP maildrop.
      valid = .false
      if isopen then
         file~Close
      isopen = .false
   
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Items                                                 |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|            Number of (potential) items in the maildrop (including deletes)|
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::METHOD Items
      expose maildrop.
      return maildrop.!items
   
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Changed                                               |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|            .true if any item in the maildrop has been deleted             |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::METHOD Changed
      expose maildrop.
      return maildrop.!changed
   
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Lines                                                 |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|            Number of lines of text in the currently selected item         |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::METHOD Lines
      expose maildrop. item
      return maildrop.item.!lines
   
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Bytes                                                 |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|            Number of octets in the currently selected item                |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::METHOD Bytes
      expose maildrop. item
      return maildrop.item.!size
   
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Valid                                                 |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|            .true if the maildrop was successfully opened                  |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::METHOD Valid
      expose valid
      return valid
   
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Select                                                |
|                                                                           |
|        Parameters:                                                        |
|            itemnum: the number of the item to be selected                 |
|                                                                           |
|        Returns:                                                           |
|            "From " divider, converted to a Return-Path (or .NIL if none)  |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::METHOD Select
      expose maildrop. item file valid
      use arg item
      if valid then
         return .NIL
         
      if item <= 0 | item > maildrop.!items then
         return .NIL
         
      if maildrop.item.!deleted then
         return .NIL
         
      xrc = file~Seek( '=' || maildrop.item.!start 'READ CHAR' )
      parse value file~LineIn with "From " username "@" hostname .
      return "Return-Path: <" || username || "@" || hostname || ">"
   
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::NextLine                                              |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|            Next line of item's header or body; .NIL after end of latter   |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::METHOD NextLine
      expose maildrop. item file
      bytepos = file~Query( 'POSITION READ CHAR' )
      if bytepos < maildrop.item.!last then
         return file~LineIn
      else
         return .NIL
      
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Size                                                  |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|            Size of entire maildrop, in octets                             |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/

   ::Method Size
      expose maildrop.
      return maildrop.!size       /* Remember, started at 1 */

/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Exists                                                |
|                                                                           |
|        Parameters:                                                        |
|            msgNo: a number to be checked as being in range for the drop   |
|                                                                           |
|        Returns:                                                           |
|            .true if message with that number COULD exist                  |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method Exists
      expose maildrop.
      use arg msgNo
      if msgNo~Datatype <> "NUM" then
         return .false
      if msgNo < 1 then
         return .false
      if msgNo > maildrop.!items then
         return .false
      return .true
      
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Deleted                                               |
|                                                                           |
|        Parameters:                                                        |
|            msgNo: a number to be checked to see whether it's already gone |
|                                                                           |
|        Returns:                                                           |
|            .true if message with that number has been deleted             |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method Deleted
      expose maildrop.
      use arg msgNo
      return maildrop.msgNo.!deleted
   
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Delete                                                |
|                                                                           |
|        Parameters:                                                        |
|            msgNo: item in maildrop to be deleted                          |
|                                                                           |
|        Returns:                                                           |
|            -1 if message already deleted                                  |
|             0 if message number cannot exist                              |
|            +1 if message deleted successfully                             |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method Delete
      expose maildrop.
      use arg msgNo
      if self~Exists( msgNo ) then
         return 0
      if self~Deleted( msgNo ) then
         return -1
      maildrop.msgNo.!deleted = .true
      maildrop.!changed = .true
      return 1

/***************************************************************************\
|                                                                           |
|     METHOD MailBox::UndeleteAll                                           |
|                                                                           |
|        Parameters:                                                        |
|                                                                           |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method UndeleteAll
      expose maildrop.
      do item = 1 to maildrop.!items
         maildrop.item.!deleted = .false
      end
      maildrop.!changed = .false
      return
      
/***************************************************************************\
|                                                                           |
|     METHOD MailBox::Rewrite                                               |
|                                                                           |
|        Parameters:                                                        |
|            tempfile: file to hold user's mailbox until writing complete   |
|                                                                           |
|        Returns:                                                           |
|                                                                           |
|        Notes:                                                             |
|                                                                           |
\***************************************************************************/
   ::Method Rewrite
      expose maildrop. file isopen
      use arg tempfile
      out = .Stream~new( tempfile )
      SAY "Writing tempfile" tempfile
      if out~Open( 'WRITE' ) <> "READY:" then 
         return .false
      xrc = 0     /* Assume all is right with the world */

      do item = 1 to maildrop.!items
         if maildrop.item.!deleted then
            Iterate     /* Ignore deleted item(s) */
         loc = file~Seek( "=" || maildrop.item.!start "READ CHAR" )
         do while loc < maildrop.item.!last
            xrc = out~LineOut( file~LineIn )
            loc = file~Query( 'Position Read Char' )
            if xrc then
               Leave item
         end
      end
      cls = out~Close
      cls = file~Close
      isopen = .false
      if xrc then do
         CALL SysFileDelete tempfile
         return .false
      end
      return .true
      
/***************************************************************************\
|                                                                           |
|     ROUTINE AddSlash                                                      |
|                                                                           |
|        Parameter:                                                         |
|           DirName: directory path, which may need a trailing '/'          |
|                                                                           |
|                                                                           |
\***************************************************************************/

::Routine AddSlash
   use arg DirName
   if Right(DirName,1) = '\' | Right(DirName,1) = '/' then
      return DirName
   else
      return DirName || '/'

